# -*- coding: utf-8 -*-

# Yokogawa.py driver for Yokogawa GS820 multi channel source measure unit
# Hannes Rotzinger, hannes.rotzinger@kit.edu 2010
# Micha Wildermuth, micha.wildermuth@kit.edu 2017
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from qkit.core.instrument_base import Instrument
from qkit import visa
import numpy as np
import time
import logging
from distutils.version import LooseVersion


class Yokogawa(Instrument):
    """
    This is the driver for the Yokogawa GS820 multi channel source measure unit.
    """
    
    def __init__(self, name, address, reset=False):
        """
        Initializes VISA communication with the instrument Yokogawa GS820.
        
        Parameters
        ----------
        name: string
            Name of the instrument (driver).
        address: string
            GPIB address for the communication with the instrument.
        reset: bool, optional
            Resets the instrument to default conditions. Default is False.
        
        Returns
        -------
        None
        
        Examples
        --------
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]
        
        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py
        
        >>> IVD = qkit.instruments.create('IVD', 'Yokogawa', address='TCPIP0::00.00.000.00::INSTR', reset=True)
        Initialized the file info database (qkit.fid) in 0.000 seconds.
        """
        self.__name__ = __name__
        # Start VISA communication
        logging.info(__name__ + ': Initializing instrument Yokogawa_GS820')
        Instrument.__init__(self, name, tags=['physical'])
        self._address = address
        self._visainstrument = visa.instrument(self._address)
        # Set termination characters (necessary for Ethernet communication)
        if LooseVersion(visa.__version__) < LooseVersion("1.5.0"):  # pyvisa 1.4
            self._visainstrument.term_chars = ''
        else:  # pyvisa 1.5
            self._visainstrument.read_termination = ''
            self._visainstrument.write_termination = ''
        # initial variables
        self._sweep_mode = 1  # IV-mode
        self._sweep_channels = (1,)
        self._measurement_modes = {0: '2-wire', 1: '4-wire'}
        self._IV_modes = {0: 'curr', 1: 'volt'}
        self._IV_units = {0: 'A', 1: 'V'}
        # dict of defaults values: defaults[<sweep_mode>][<channel>][<parameter>][<value>]
        self._defaults = {0: [{self.set_measurement_mode: 0,
                               self.set_bias_mode: 1,
                               self.set_sense_mode: 1,
                               self.set_bias_range: -1,
                               self.set_sense_range: -1,
                               self.set_bias_delay: 200e-6,
                               self.set_sense_delay: 15e-6,
                               self.set_sense_nplc: 1,
                               self.set_sense_average: 1,
                               self.set_sense_autozero: 0},
                              {self.set_measurement_mode: 0,
                               self.set_bias_mode: 0,
                               self.set_sense_mode: 1,
                               self.set_bias_range: -1,
                               self.set_sense_range: -1,
                               self.set_bias_delay: 200e-6,
                               self.set_sense_delay: 15e-6,
                               self.set_sense_nplc: 1,
                               self.set_sense_average: 1,
                               self.set_sense_autozero: 0}],
                          1: [{self.set_measurement_mode: 1,
                               self.set_bias_mode: 0,
                               self.set_sense_mode: 1,
                               self.set_bias_trigger: '''str('sens')''',
                               self.set_sense_trigger: '''str('sour')''',
                               self.set_bias_range: -1,
                               self.set_sense_range: -1,
                               self.set_bias_delay: 200e-6,
                               self.set_sense_delay: 15e-6,
                               self.set_sense_nplc: 1,
                               self.set_sense_average: 1,
                               self.set_sense_autozero: 0}],
                          2: [{self.set_measurement_mode: 1,
                               self.set_bias_mode: 1,
                               self.set_sense_mode: 0,
                               self.set_bias_trigger: '''str('sens')''',
                               self.set_sense_trigger: '''str('sour')''',
                               self.set_bias_range: -1,
                               self.set_sense_range: -1,
                               self.set_bias_delay: 200e-6,
                               self.set_sense_delay: 15e-6,
                               self.set_sense_nplc: 1,
                               self.set_sense_average: 1,
                               self.set_sense_autozero: 0}]}
        self._default_str = 'default parameters'
        self.__set_defaults_docstring()
        # condition code registers
        self.bias_ccr = [('EOS1', 'CH1 End of Sweep'),
                         ('RDY1', 'CH1 Ready for Sweep'),
                         ('LL01', 'CH1 Low Limiting'),
                         ('LHI1', 'CH1 High Limiting'),
                         ('TRP1', 'CH1 Tripped'),
                         ('EMR1', 'CH1 Emergency (Temperature/Current over)'),
                         ('', ''),
                         ('', ''),
                         ('EOS2', 'CH2 End of Sweep'),
                         ('RDY2', 'CH2 Ready for Sweep'),
                         ('LL02', 'CH2 Low Limiting'),
                         ('LHI2', 'CH2 High Limiting'),
                         ('TRP2', 'CH2 Tripped'),
                         ('EMR2', 'CH2 Emergency (Temperature/Current over)'),
                         ('ILC', 'Inter Locking'),
                         ('SSB', 'Start Sampling Error')]
        self.sense_ccr = [('EOM1', 'CH1 End of Measure'),
                          ('', ''),
                          ('CLO1', 'CH1 Compare result is Low'),
                          ('CHI1', 'CH1 Compare result is High'),
                          ('', ''),
                          ('OVR1', 'CH1 Over Range'),
                          ('', ''),
                          ('', ''),
                          ('EOM2', 'CH2 End of Measure'),
                          ('', ''),
                          ('CLO2', 'CH2 Compare result is Low'),
                          ('CHI2', 'CH1 Compare result is High'),
                          ('', ''),
                          ('OVR2', 'CH2 Over Range'),
                          ('EOT', 'End of Trace'),
                          ('TSE', 'Trigger Sampling Error')]
        # error messages
        self.err_msg = {-101: '''Invalid_character: Check whether invalid characters such as $ or & are used in the command header or parameters.''',
                        -102: '''Syntax_error: Check that the syntax is correct.''',
                        -103: '''Invalid separator: Check the use of the separator (comma).''',
                        -106: '''Parameter not allowed: Check the command and the number of parameters.''',
                        -107: '''Missing parameter: Check the command and the number of parameters.''',
                        -112: '''Program mnemonic too long: Check the command mnemonic.''',
                        -113: '''Undefined header: Check the command mnemonic.''',
                        -121: '''Invalid character in number: Check that the notation of the numeric parameter is correct (for example, binary notation should not contain characters other than 0 and 1)''',
                        -122: '''Header suffix out of range: Check whether the numeric suffix of the command header is correct.''',
                        -123: '''Exponent too large: Check whether the exponent is within the range of -127 to 127.''',
                        -124: '''Too many digits: Check that the number of digits in the value does not exceed 255.''',
                        -128: '''Numeric data not allowed: Check the parameter format.''',
                        -131: '''Invalid suffix: Check the unit that can be used for the parameter.''',
                        -138: '''Suffix not allowed: Check the parameter format.''',
                        -141: '''Invalid character data: Check the character data that can be used for the parameter.''',
                        -148: '''Character data not allowed: Check the command and parameter format.''',
                        -150: '''String data error: Check that the closing quotation mark (" or ') for a string is available.''',
                        -151: '''Invalid string data: Check that the string parameter is in the correct format.''',
                        -158: '''String data not allowed: Check the command and parameter format.''',
                        -161: '''Invalid block data: Check that the block data is in the correct format.''',
                        -168: '''Block data not allowed: Check the command and parameter format.''',
                        -178: '''Expression data not allowed: Check the command and parameter format.''',
                        -222: '''Data out of range: Check the selectable range of the parameter. If the command can use MINimum and MAXimum as its parameter, the range can also be queried.''',
                        -256: '''Filename not found: Check that the file exists. You can also use the CATalog? command to query the list of files.''',
                        -285: '''Program syntax error: Check that the sweep pattern file is in the correct format.''',
                        -350: '''Queue overflow: Read the error using :SYSTem:ERRor? or clear the error queue using *CLS.''',
                        -361: '''Parity error: Check that the communication settings on the GS820 and PC match. If the settings are correct, check the cable, and lower the baud rate.''',
                        -362: '''Framing error: Check that the communication settings on the GS820 and PC match. If the settings are correct, check the cable, and lower the baud rate.''',
                        -363: '''Input buffer overrun: Set the handshaking to a setting other than OFF. Lower the baud rate.''',
                        -410: '''Query INTERRUPTED: Check transmission/reception procedure.''',
                        -420: '''Query UNTERMINATED: Check transmission/reception procedure.''',
                        -430: '''Query DEADLOCK: Keep the length of a program message less than or equal to 64 KB.''',
                        +101: '''Too complex expression: Keep the total number of constants, variables, and operators in a MATH definition less than or equal to 256.''',
                        +102: '''Math file syntax error: Check that the syntax of the MATH definition file is correct.''',
                        +103: '''Too large file error: Keep MATH definition files less than 4 KB in size.''',
                        +104: '''Illegal file error: Download the file for updating the system firmware again.''',
                        +105: '''No slave SMU found: Check that the connection between the master and slave units is correct.''',
                        +200: '''Sweep stopped because of the setting change: Stop the sweep operation before changing the settings.''',
                        +202: '''Interlocking: Release the interlock, and then turn the output ON.''',
                        +203: '''Cannot relay on in hardware abnormal: Check whether the temperature inside the case is okay.''',
                        +204: '''Hardware input abnormal error: Connect a load within the specifications.''',
                        +205: '''Analog busy: Change the settings after the calibration or self-test is completed.''',
                        +206: '''Low battery: Request to have the battery replaced, because the time stamp when creating files will not be correct.''',
                        +207: '''Power line frequency measure failure: Directly set the line frequency.''',
                        +304: '''Cannot change setting in auto measure function: If you want to change the measurement function, select a measurement mode other than auto function.'''}
        # reset
        if reset:
            self.reset()
        else:
            self.get_all()
        return
    
    def _write(self, cmd):
        """
        Sends a visa command <cmd>, waits until "operation complete" and raises eventual errors of the Device.
        
        Parameters
        ----------
        cmd: str
            Command that is send to the instrument via pyvisa and NI-VISA backend.
        
        Returns
        -------
        None
        """
        self._visainstrument.write(cmd)
        while not bool(int(self._visainstrument.query('*OPC?'))):
            time.sleep(1e-6)
        self._raise_error()
        return
    
    def _ask(self, cmd):
        """
        Sends a visa command <cmd>, waits until "operation complete", raises eventual errors of the Device and returns the read answer <ans>.
        
        Parameters
        ----------
        cmd: str
            Command that is send to the instrument via pyvisa and NI-VISA backend.
        
        Returns
        -------
        answer: str
            Answer that is returned at query after the sent <cmd>.
        """
        if '?' in cmd:
            ans = self._visainstrument.query(cmd).rstrip()
        else:
            ans = self._visainstrument.query('{:s}?'.format(cmd)).rstrip()
        while not bool(int(self._visainstrument.query('*OPC?'))):
            time.sleep(1e-6)
        self._raise_error()
        return ans
    
    def set_measurement_mode(self, mode, channel=1):
        """
        Sets measurement mode (wiring system) of channel <channel> to <mode>.
        
        Parameters
        ----------
        mode: int
            State of the measurement sense mode. Must be 0 (2-wire) or 1 (4-wire).
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:REMote 1|0|ON|OFF
        try:
            logging.debug('{!s}: Set measurement mode of channel {:d} to {:d}'.format(__name__, channel, mode))
            self._write(':chan{:d}:sens:rem {:d}'.format(channel, mode))
        except Exception as e:
            logging.error('{!s}: Cannot set measurement mode of channel {!s} to {!s}'.format(__name__, channel, mode))
            raise type(e)('{!s}: Cannot set measurement mode of channel {!s} to {!s}\n{!s}'.format(__name__, channel, mode, e))
        return
    
    def get_measurement_mode(self, channel=1):
        """
        Gets measurement mode (wiring system) <mode> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        mode: int
            State of the measurement sense mode. Meanings are 0 (2-wire) and 1 (4-wire).
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:REMote 1|0|ON|OFF
        try:
            logging.debug('{!s}: Get measurement mode of channel {:d}'.format(__name__, channel))
            return int(self._ask(':chan{:d}:sens:rem'.format(channel)))
        except Exception as e:
            logging.error('{!s}: Cannot get measurement mode of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get measurement mode of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_sync(self, val):
        """
        Sets the interchannel synchronization to <val>. The first channel is always the "master", the second the "slave".
        
        Parameters
        ----------
        val: bool
			Inter-channel synchronization mode specifies whether two channels are to be operated in sync.
		
        Returns
        -------
        None
        """
        # Corresponding Command: :SYNChronize:CHANnel 1|0|ON|OFF
        try:
            logging.debug('{!s}: Set interchannel synchronization to {!r}'.format(__name__, val))
            self._write(':sync:chan {:d}'.format(val))
        except Exception as e:
            logging.error('{!s}: Cannot set interchannel synchronization to {!s}'.format(__name__, val))
            raise type(e)('{!s}: Cannot set interchannel synchronization to {!s}\n{!s}'.format(__name__, val, e))
        return
    
    def get_sync(self):
        """
        Gets the interchannel synchronization.
        
        Parameters
        ----------
        None
		
        Returns
        -------
        val: bool
			Inter-channel synchronization mode specifies whether two channels are to be operated in sync.
        """
        # Corresponding Command: :SYNChronize:CHANnel 1|0|ON|OFF
        try:
            logging.debug('{!s}: Get interchannel synchronization'.format(__name__))
            return bool(int(self._ask(':sync:chan')))
        except Exception as e:
            logging.error('{!s}: Cannot get interchannel synchronization'.format(__name__))
            raise type(e)('{!s}: Cannot get interchannel synchronization\n{!s}'.format(__name__, e))
    
    def set_bias_mode(self, mode, channel=1):
        """
        Sets bias mode of channel <channel> to mode <mode>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        mode: int
            Bias mode. Must be 0 (current) or 1 (voltage).
        
        Returns
        -------
        None
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce:FUNCtion VOLTage|CURRent
        try:
            logging.debug('{!s}: Set bias mode of channel {:d} to {:d}'.format(__name__, channel, mode))
            self._write(':chan{:d}:sour:func {:s}'.format(channel, self._IV_modes[mode]))
            self.set_bias_range(val=-1, channel=channel)
        except Exception as e:
            logging.error('{!s}: Cannot set bias mode of channel {!s} to {!s}'.format(__name__, channel, mode))
            raise type(e)('{!s}: Cannot set bias mode of channel {!s} to {!s}\n{!s}'.format(__name__, channel, mode, e))
        return
    
    def get_bias_mode(self, channel=1):
        """
        Gets bias mode <mode> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        mode: int
            Bias mode. Meanings are 0 (current) and 1 (voltage).
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce:FUNCtion VOLTage|CURRent
        try:
            logging.debug('{!s}: Get bias mode of channel {:d}'.format(__name__, channel))
            return int(list(self._IV_modes.keys())[list(self._IV_modes.values()).index(self._ask(':chan{:d}:sour:func'.format(channel)).lower())])
        except Exception as e:
            logging.error('{!s}: Cannot get bias mode of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get bias mode of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_sense_mode(self, mode, channel=1):
        """
        Sets sense mode of channel <channel> to mode <mode>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        mode: int
            Sense mode. Must be 0 (current) or 1 (voltage).
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:FUNCtion VOLTage|CURRent
        try:
            logging.debug('{!s}: Set sense mode of channel {:d} to {:d}'.format(__name__, channel, mode))
            self._write(':chan{:d}:sens:func {:s}'.format(channel, self._IV_modes[mode]))
            self.set_sense_range(val=-1, channel=channel)
        except Exception as e:
            logging.error('{!s}: Cannot set sense mode of channel {!s} to {!s}'.format(__name__, channel, mode))
            raise type(e)('{!s}: Cannot set sense mode of channel {!s} to {!s}\n{!s}'.format(__name__, channel, mode, e))
        return
    
    def get_sense_mode(self, channel=1):
        """
        Gets sense mode <mode> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        mode: int
            Sense mode. Meanings are 0 (current) and 1 (voltage).
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:FUNCtion VOLTage|CURRent
        try:
            logging.debug('{!s}: Get sense mode of channel {:d}'.format(__name__, channel))
            return int(list(self._IV_modes.keys())[list(self._IV_modes.values()).index(self._ask(':chan{:d}:sens:func'.format(channel)).lower())])
        except Exception as e:
            logging.error('{!s}: Cannot get sense mode of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sense mode of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_bias_range(self, val, channel=1):
        """
        Sets bias range of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Bias range. Possible values are -1 (auto), for currents 200nA, 2uA, 20uA, 200uA, 2mA, 20mA, 200mA, 1 A, 3A and for voltages 200mV, 2V, 7V, 18V.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce[:VOLTage]:RANGe <voltage>|MINimum|MAXimum|UP|DOWN 
        # Corresponding Command: [:CHANnel<n>]:SOURce[:CURRent]:RANGe <current>|MINimum|MAXimum|UP|DOWN 
        try:
            logging.debug('{!s}: Set bias range of channel {:d} to {:f}'.format(__name__, channel, val))
            if val == -1:
                self._write(':chan{:d}:sour:rang:auto 1'.format(channel))
            else:
                self._write(':chan{:d}:sour:rang {:f}'.format(channel, val))
        except Exception as e:
            logging.error('{!s}: Cannot set bias range of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set bias range of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def get_bias_range(self, channel=1):
        """
        Gets bias mode <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Bias range.
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce[:VOLTage]:RANGe <voltage>|MINimum|MAXimum|UP|DOWN 
        # Corresponding Command: [:CHANnel<n>]:SOURce[:CURRent]:RANGe <current>|MINimum|MAXimum|UP|DOWN 
        try:
            logging.debug('{!s}: Get bias range of channel {:d}'.format(__name__, channel))
            return float(self._ask(':chan{:d}:sour:rang'.format(channel)))
        except Exception as e:
            logging.error('{!s}: Cannot get bias range of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get bias range of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_sense_range(self, val, channel=1):
        """
        Sets sense range of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Sense range. Possible values are -1 (auto), for currents 200nA, 2uA, 20uA, 200uA, 2mA, 20mA, 200mA, 1 A, 3A and for voltages 200mV, 2V, 7V, 18V.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe[:VOLTage]:RANGe <voltage>|MINimum|MAXimum|UP|DOWN
        # Corresponding Command: [:CHANnel<n>]:SENSe[:CURRent]:RANGe <current>|MINimum|MAXimum|UP|DOWN
        try:
            logging.debug('{!s}: Set sense range of channel {:d} to {:f}'.format(__name__, channel, val))
            if val == -1:
                self._write(':chan{:d}:sens:rang:auto 1'.format(channel))
            else:
                self._write(':chan{:d}:sens:rang {:f}'.format(channel, val))
        except Exception as e:
            logging.error('{!s}: Cannot set sense range of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set sense range of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def get_sense_range(self, channel=1):
        """
        Gets sense mode <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Sense range.
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe[:VOLTage]:RANGe <voltage>|MINimum|MAXimum|UP|DOWN
        # Corresponding Command: [:CHANnel<n>]:SENSe[:CURRent]:RANGe <current>|MINimum|MAXimum|UP|DOWN
        try:
            logging.debug('{!s}: Get sense range of channel {:d}'.format(__name__, channel))
            return float(self._ask(':chan{:d}:sens:rang'.format(channel)))
        except Exception as e:
            logging.error('{!s}: Cannot get sense range of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sense range of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_bias_trigger(self, mode, channel=1, **val):
        """
        Sets bias trigger mode of channel <channel> to <mode> and optional value <val>.
        
        Parameters
        ----------
        mode: str
			Bias trigger mode. Must be 'ext' (external), 'aux' (auxiliary), 'tim1' (timer1), 'tim2' (timer2) or 'sens' (sense).
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        time: float, optional
			Time trigger period if <mode> is 'timer'. Must be in [100µs, 3600s].
		pol: int, optional
			Polarity of auxiliary trigger if <mode> is 'aux'. Must be 0 (falling edge) or 1 (rising edge).
		
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce:TRIGger EXTernal|AUXiliary|TIMer1|TIMer2|SENSe
        # Corresponding Command: [:CHANnel<n>]:SOURce:TRIGger:AUXiliary:POLarity NORMal|INVerted
        # Corresponding Command: :TRIGger:TIMer1 <time>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Set bias trigger of channel {:d} to {:s}'.format(__name__, channel, mode))
            self._write(':chan{:d}:sour:trig {:s}'.format(channel, mode))
            if 'pol' in val:
                cmd = {0: 'norm', 1: 'inv'}
                self._write(':chan{:d}:sour:trig:aux:pol {:s}'.format(channel, cmd[val.get('pol', 0)]))
            if 'time' in val:
                self._write(':trig:{:s} {:f}'.format(mode, val.get('time', 50e-3)))
        except Exception as e:
            logging.error('{!s}: Cannot set bias trigger of channel {!s} to {!s}{!s}'.format(__name__, channel, mode, val))
            raise type(e)('{!s}: Cannot set bias trigger of channel {!s} to {!s}{!s}\n{!s}'.format(__name__, channel, mode, val, e))
        return
    
    def get_bias_trigger(self, channel=1):
        """
        Gets bias trigger of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        mode: str
			Bias trigger mode. Meanings are 'ext' (external), 'aux' (auxiliary), 'tim1' (timer1), 'tim2' (timer2) and 'sens' (sense).
        time: float, optional
			Time trigger period if <mode> is 'timer'.
		pol: int, optional
			Polarity of auxiliary trigger if <mode> is 'aux'. Meanings are 0 (falling edge) and 1 (rising edge).
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce:TRIGger EXTernal|AUXiliary|TIMer1|TIMer2|SENSe
        # Corresponding Command: [:CHANnel<n>]:SOURce:TRIGger:AUXiliary:POLarity NORMal|INVerted
        # Corresponding Command: :TRIGger:TIMer1 <time>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Get bias trigger of channel {:d}'.format(__name__, channel))
            mode = str(self._ask(':chan{:d}:sour:trig'.format(channel)).lower())
            if mode[:3] == 'tim':
                val = float(self._ask(':trig:{:s}'.format(mode)))
                return mode, val
            elif mode == 'aux':
                val = int(float(self._ask(':chan{:d}:sour:trig:aux:pol'.format(channel))))
                return mode, val
            else:
                return mode
        except Exception as e:
            logging.error('{!s}: Cannot get bias trigger of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get bias trigger of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_sense_trigger(self, mode, channel=1, **val):
        """
        Sets sense trigger mode of channel <channel> to <mode> and optional value <val>. If <mode> is 'timer' it can be set to <time>
        
        Parameters
        ----------
        mode: str
			Sense trigger mode. Must be 'sour' (source), 'swe' (sweep), 'aux' (auxiliary), 'tim1' (timer1), 'tim2' (timer2) or 'imm' (immediate).
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        time: float, optional
			Time trigger period. Must be in [100µs, 3600s].
		pol: int, optional
			Polarity of auxiliary trigger if <mode> is 'aux'. Must be 0 (falling edge) or 1 (rising edge).
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:TRIGger SOURce|SWEep|AUXiliary|TIMer1|TIMer2|IMMediate
        # Corresponding Command: [:CHANnel<n>]:SENSe:TRIGger:AUXiliary:POLarity NORMal|INVerted
        # Corresponding Command: :TRIGger:TIMer1 <time>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Set sense trigger of channel {:d} to {:s}'.format(__name__, channel, mode))
            self._write(':chan{:d}:sens:trig {:s}'.format(channel, mode))
            if 'time' in val:
                self._write(':trig:{:s} {:f}'.format(mode, val.get('time', 50e-3)))
            elif 'pol' in val:
                cmd = {0: 'norm', 1: 'inv'}
                self._write(':chan{:d}:sens:trig:aux:pol {:s}'.format(channel, cmd[val.get('pol', 0)]))
        except Exception as e:
            logging.error('{!s}: Cannot set sense trigger of channel {!s} to {!s}{!s}'.format(__name__, channel, mode, val))
            raise type(e)('{!s}: Cannot set sense trigger of channel {!s} to {!s}{!s}\n{!s}'.format(__name__, channel, mode, val, e))
        return
    
    def get_sense_trigger(self, channel=1):
        """
        Gets sense trigger of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        mode: str
			Sense trigger mode. Meanings are 'sour' (source), 'swe' (sweep), 'aux' (auxiliary), 'tim1' (timer1), 'tim2' (timer2) or 'imm' (immediate).
        time: float, optional
			Time trigger period.
		pol: int, optional
			Polarity of auxiliary trigger if <mode> is 'aux'. Meanings are 0 (falling edge) and 1 (rising edge).
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:TRIGger SOURce|SWEep|AUXiliary|TIMer1|TIMer2|IMMediate
        # Corresponding Command: [:CHANnel<n>]:SENSe:TRIGger:AUXiliary:POLarity NORMal|INVerted
        # Corresponding Command: :TRIGger:TIMer1 <time>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Get sense trigger of channel {:d}'.format(__name__, channel))
            mode = str(self._ask(':chan{:d}:sens:trig'.format(channel)).lower())
            if mode[:3] == 'tim':
                val = float(self._ask(':trig:{:s}'.format(mode)))
                return mode, val
            elif mode == 'aux':
                val = int(float(self._ask(':chan{:d}:sens:trig:aux:pol'.format(channel))))
                return mode, val
            else:
                return mode
        except Exception as e:
            logging.error('{!s}: Cannot get sense trigger of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sense trigger of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_bias_delay(self, val, channel=1):
        """
        Sets bias delay of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Bias delay with respect to the bias trigger. Must be are in [15µs, 3600s].
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce:DELay <time>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Set bias delay of channel {:d} to {:f}'.format(__name__, channel, val))
            self._write(':chan{:d}:sour:del {:f}'.format(channel, val))
        except Exception as e:
            logging.error('{!s}: Cannot set bias delay of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set bias delay of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def get_bias_delay(self, channel=1):
        """
        Gets bias delay <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Bias delay with respect to the bias trigger.
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce:DELay <time>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Get bias delay of channel {:d}'.format(__name__, channel))
            return float(self._ask(':chan{:d}:sour:del'.format(channel)))
        except Exception as e:
            logging.error('{!s}: Cannot get bias delay of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get bias delay of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_sense_delay(self, val, channel=1, **kwargs):
        """
        Sets sense delay of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Sense delay with respect to the sense trigger. Must be are in [15µs, 3600s].
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce:DELay <time>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Set sense delay of channel {:d} to {:f}'.format(__name__, channel, val))
            self._write(':chan{:d}:sens:del {:f}'.format(channel, val))
        except Exception as e:
            logging.error('{!s}: Cannot set sense delay of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set sense delay of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def get_sense_delay(self, channel=1):
        """
        Gets sense delay <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Sense delay with respect to the sense trigger.
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:DELay <time>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Get sense delay of channel {:d}'.format(__name__, channel))
            return float(self._ask(':chan{:d}:sens:del'.format(channel)))
        except Exception as e:
            logging.error('{!s}: Cannot get sense delay of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sense delay of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_sense_average(self, val, channel=1, **kwargs):
        """
        Sets sense average of channel <channel> to <val>.
        If <val> is less than 1 average status is turned off, but the set value remains.
        
        Parameters
        ----------
        val: int
            Number of measured readings that are required to yield one filtered measurement. Must be in [1, 100].
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:AVERage[:STATe] 1|0|ON|OFF
        # Corresponding Command: [:CHANnel<n>]:SENSe:AVERage:COUNt <integer>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Set sense average of channel {:d} to {:d}'.format(__name__, channel, val))
            status = not(.5*(1-np.sign(val-1)))  # equals Heaviside(1-<val>) --> turns on for <val> >= 2
            self._write(':chan{:d}:sens:aver:stat {:d}'.format(channel, status))
            if status:
                self._write(':chan{:d}:sens:aver:coun {:d}'.format(channel, val))
        except Exception as e:
            logging.error('{!s}: Cannot set sense average of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set sense average of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def get_sense_average(self, channel=1):
        """
        Gets sense average status <status> and value <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        status: bool
            Status if the measurement filter is enabled.
        val: int
            Number of measured readings that are required to yield one filtered measurement.
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:AVERage[:STATe] 1|0|ON|OFF
        # Corresponding Command: [:CHANnel<n>]:SENSe:AVERage:COUNt <integer>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Get sense average of channel {:d}'.format(__name__, channel))
            status = bool(int(self._ask(':chan{:d}:sens:aver:stat'.format(channel))))
            val = int(self._ask(':chan{:d}:sens:aver:coun'.format(channel)))
            return status, val
        except Exception as e:
            logging.error('{!s}: Cannot get sense average of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sense average of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_plc(self, plc):
        """
        Sets power line cycle (PLC) to <val>.
        
        Parameters
        ----------
        plc: int
            Power line frequency setting used for NPLC calculations. Possible values are -1 (auto) | 50 | 60.
        
        Returns
        -------
		None
        """
        # Corresponding Command: :SYSTem:LFRequency 50|60
        # Corresponding Command: :SYSTem:LFRequency:AUTO 1|0|ON|OFF
        try:
            logging.debug('{!s}: Set PLC to {!s}'.format(__name__, plc))
            cmd = {-1: ':auto 1', 50: ' 50', 60: ' 60'}
            self._write('syst:lfr{:s}'.format(cmd.get(int(plc), cmd[-1])))
        except Exception as e:
            logging.error('{!s}: Cannot set PLC to {!s}'.format(__name__, plc))
            raise type(e)('{!s}: Cannot set PLC to {!s}\n{!s}'.format(__name__, plc, e))
        return
    
    def get_plc(self):
        """
        Gets power line cycle (PLC) <val>.
        
        Parameters
        ----------
		None
        
        Returns
        -------
        plc: int
            Power line frequency setting used for NPLC calculations.
        """
        # Corresponding Command: :SYSTem:LFRequency 50|60
        # Corresponding Command: :SYSTem:LFRequency:AUTO 1|0|ON|OFF
        try:
            logging.debug('{!s}: Get PLC'.format(__name__))
            return int(self._ask('syst:lfr'))
        except Exception as e:
            logging.error('{!s}: Cannot get PLC'.format(__name__))
            raise type(e)('{!s}: Cannot get PLC\n{!s}'.format(__name__, e))
    
    def set_sense_nplc(self, val, channel=1):
        """
        Sets sense nplc (number of power line cycle) of channel <channel> with the <val>-fold of one power line cycle.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        val: float
            Integration aperture for measurements. Must be in [1, 25].
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:NPLC <real number>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Set sense nplc of channel {:d} to {:2.0f} PLC'.format(__name__, channel, val))
            self._write(':chan{:d}:sens:nplc {:2.0f}'.format(channel, val))
        except Exception as e:
            logging.error('{!s}: Cannot set sense nplc of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set sense nplc of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def get_sense_nplc(self, channel=1):
        """
        Gets sense nplc (number of power line cycle) <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Integration aperture for measurements.
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:NPLC <real number>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Get sense nplc of channel {:d}'.format(__name__, channel))
            return float(self._ask(':chan{:d}:sens:nplc'.format(channel)))
        except Exception as e:
            logging.error('{!s}: Cannot get sense nplc of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sense nplc of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_sense_autozero(self, val, channel=1):
        """
        Sets autozero of channel <channel> to <val>. Note that "on" means that the internal zero point is measured for each measurement, wherefore the measurement takes approximately twice as long as when the auto zero function is "off"
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        val: int
            Status of the internal reference measurements (autozero) of the source-measure unit. Possible values are 0 (off), 1 (on) or 2 (once).
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:ZERo:AUTO 1|0|ON|OFF (at each measurement)
        # Corresponding Command: [:CHANnel<n>]:SENSe:ZERo:EXECute (once)
        try:
            logging.debug('{!s}: Set sense autozero of channel {:d} to {:d}'.format(__name__, channel, val))
            if val == 2:
                self._visainstrument.write(':chan{:d}:sens:zero:exec'.format(channel))
                time.sleep(2.)  # tests yield random value t > 1.65 to be necessary
                self._wait_for_OPC()
                self._raise_error()
            else:
                self._write(':chan{:d}:sens:zero:auto {:d}'.format(channel, val))
        except Exception as e:
            logging.error('{!s}: Cannot set sense autozero of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set sense autozero of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def get_sense_autozero(self, channel=1):
        """
        Gets sense autozero <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: int
            Status of the internal reference measurements (autozero) of the source-measure unit. Meanings are 0 (off) and 1 (on).
        """
        # Corresponding Command: [:CHANnel<n>]:SENSe:ZERo:AUTO 1|0|ON|OFF
        try:
            logging.debug('{!s}: Get sense autozero of channel {:d}'.format(__name__, channel))
            return bool(self._ask(':chan{:d}:sens:zero:auto'.format(channel)))
        except Exception as e:
            logging.error('{!s}: Cannot get sense autozero of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sense autozero of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_status(self, status, channel=1):
        """
        Sets output status of channel <channel> to <status>.
        
        Parameters
        ----------
        status: int
            Output status. Possible values are 0 (off) or 1 (on).
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:OUTput:STATus 1|0|ON|OFF
        try:
            logging.debug('{!s}: Set output status of channel {:d} to {!r}'.format(__name__, channel, status))
            self._write(':chan{:d}:outp:stat {:d}'.format(channel, status))
        except Exception as e:
            logging.error('{!s}: Cannot set output status of channel {!s} to {!s}'.format(__name__, channel, status))
            raise type(e)('{!s}: Cannot set output status of channel {!s} to {!s}\n{!s}'.format(__name__, channel, status, e))
        return
    
    def get_status(self, channel=1):
        """
        Gets output status <status> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        status: int
            Output status. Meanings are 0 (off) and 1 (on).
        """
        # Corresponding Command: [:CHANnel<n>]:OUTput:STATus 1|0|ON|OFF
        try:
            logging.debug('{!s}: Get output status of channel {:d}'.format(__name__, channel))
            return bool(int(self._ask(':chan{:d}:outp:stat'.format(channel))))
        except Exception as e:
            logging.error('{!s}: Cannot get output status of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get output status of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_bias_value(self, val, channel=1):
        """
        Sets bias value of channel <channel> to value <val>.
        
        Parameters
        ----------
        val: float
            Bias value.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce[:VOLTage]:LEVel <voltage>|MINimum|MAXimum
        # Corresponding Command: [:CHANnel<n>]:SOURce[:CURRent]:LEVel <current>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Set bias value of channel {:d} to {:f}'.format(__name__, channel, val))
            self._write(':chan{:d}:sour:lev {:g}'.format(channel, val))  # necessary to cast as scientific float! (otherwise only >= 1e-6 possible)
        except Exception as e:
            logging.error('{!s}: Cannot set bias value of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set bias value of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def get_bias_value(self, channel=1):
        """
        Gets bias value <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Bias value.
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce[:VOLTage]:LEVel <voltage>|MINimum|MAXimum
        # Corresponding Command: [:CHANnel<n>]:SOURce[:CURRent]:LEVel <current>|MINimum|MAXimum
        try:
            logging.debug('{!s}: Get bias value of channel {:d}'.format(__name__, channel))
            return float(self._ask(':chan{:d}:sour:lev'.format(channel)))
        except Exception as e:
            logging.error('{!s}: Cannot get bias value of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get bias value of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def get_sense_value(self, channel=1):
        """
        Gets sense value of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Sense value.
        """
        # Corresponding Command: [:CHANnel<n>]:MEASure?
        try:
            logging.debug('{!s}: Get sense value of channel {:d}'.format(__name__, channel))
            return float(self._ask(':chan{:d}:meas'.format(channel)))
        except Exception as e:
            logging.error('{!s}: Cannot get sense value of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sense value of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_voltage(self, val, channel=1):
        """
        Sets voltage value of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Bias voltage value.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        if self.get_bias_mode(channel):  # 1 (voltage bias)
            return self.set_bias_value(val, channel)
        elif not self.get_bias_mode(channel):  # 0 (current bias)
            logging.error('{!s}: Cannot set voltage value of channel {!s} in the current bias'.format(__name__, channel))
            raise SystemError('{!s}: Cannot set voltage value of channel {!s} in the current bias'.format(__name__, channel))
        return
    
    def get_voltage(self, channel=1, **kwargs):
        """
        Gets voltage value <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Sense voltage value.
        """
        if self.get_bias_mode(channel):  # 1 (voltage bias)
            return self.get_bias_value(channel)
        elif self.get_sense_mode(channel):  # 1 (voltage sense)
            return self.get_sense_value(channel)
        else:
            logging.error('{!s}: Cannot get voltage value of channel {!s}: neither bias nor sense in voltage mode'.format(__name__, channel))
            raise SystemError('{!s}: Cannot get voltage value of channel {!s}: neither bias nor sense in voltage mode'.format(__name__, channel))
    
    def set_current(self, val, channel=1):
        """
        Sets current value of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Bias current value.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        if not self.get_bias_mode(channel):  # 0 (current bias)
            return self.set_bias_value(val, channel)
        elif self.get_bias_mode(channel):  # 1 (voltage bias)
            logging.error('{!s}: Cannot set current value of channel {!s} in the voltage bias'.format(__name__, channel))
            raise SystemError('{!s}: Cannot set current value of channel {!s} in the voltage bias'.format(__name__, channel))
        return
    
    def get_current(self, channel=1):
        """
        Gets current value <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Sense current value.
        """
        if not self.get_bias_mode(channel):  # 0 (current bias)
            return self.get_bias_value(channel)
        elif not self.get_sense_mode(channel):  # 0 (current sense)
            return self.get_sense_value(channel)
        else:
            logging.error('{!s}: Cannot get current value of channel {!s}: neither bias nor sense in current mode'.format(__name__, channel))
            raise SystemError('{!s}: Cannot get current value of channel {!s}: neither bias nor sense in current mode'.format(__name__, channel))
    
    def get_IV(self, channel=1):
        """
        Gets both current value <I_val> and voltage value <V_val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        I_val: float
            Sense current value.
        V_val: float
            Sense voltage value.
        """
        return self.get_current(channel=channel), self.get_voltage(channel=channel)
    
    def ramp_bias(self, stop, step, step_time=0.1, channel=1):
        """
        Ramps bias value of channel <channel> from recent value to stop value <stop> with step size <step> and step time <step_time>.
        
        Parameters
        ----------
        stop: float
            Stop value
        step: float
            Step size.
        step_time: float
            Sleep time between staircase ramp.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1. (default)
        
        Returns
        -------
		None
        """
        start = self.get_bias_value(channel=channel)
        if stop < start:
            step = -step
        for val in np.arange(start, stop, step)+step:
            self.set_bias_value(val, channel=channel)
            print('{:f}{:s}'.format(val, self._IV_units[self.get_bias_mode()]), end='\r')
            time.sleep(step_time)
        return
    
    def ramp_voltage(self, stop, step, step_time=0.1, channel=1):
        """
        Ramps voltage of channel <channel> from recent value to stop value <stop> with step size <step> and step time <step_time> according to bias_mode.
        
        Parameters
        ----------
        stop: float
            Stop voltage value
        step: float
            Step voltage size.
        step_time: float
            Sleep time between voltage staircase ramp.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1. (default)
        
        Returns
        -------
		None
        """
        if self.get_bias_mode(channel=channel):  # 1 (voltage bias)
            return self.ramp_bias(stop=stop, step=step, step_time=step_time, channel=channel)
        elif not self.get_bias_mode(channel=channel):  # 0 (current bias)
            logging.error(__name__ + ': Cannot set voltage value of channel {!s}: in the current bias'.format(channel))
            return
    
    def ramp_current(self, stop, step, step_time=0.1, channel=1):
        """
        Ramps current of channel <channel> from recent value to stop value <stop> with step size <step> and step time <step_time> according to bias_mode.
        
        Parameters
        ----------
        stop: float
            Stop current value
        step: float
            Step current size.
        step_time: float
            Sleep time between current staircase ramp.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1. (default)
        
        Returns
        -------
		None
        """
        if not self.get_bias_mode(channel=channel):  # 0 (current bias)
            return self.ramp_bias(stop=stop, step=step, step_time=step_time, channel=channel)
        elif self.get_bias_mode(channel=channel):  # 1 (voltage bias)
            logging.error(__name__ + ': Cannot ramp current value of channel {!s}: in the voltage bias'.format(channel))
            return
    
    def set_sweep_mode(self, mode=1, *channels):
        """
        Sets an variable to decide weather
         * voltage is both applied and measured (VV-mode),
         * current is applied and voltage is measured (IV-mode) or
         * voltage is applied and current is measured (VI-mode)
        and channels <channels> that are used during sweeps.
        
        Parameters
        ----------
        mode: int
            Sweep mode denoting bias and sense modes. Possible values are 0 (VV-mode), 1 (IV-mode) or 2 (VI-mode). Default is 1
        channels: int or tuple(int)
            Number of channel of usage. Must be 1 or 2 for IV-mode and VI-mode or a tuple containing both channels for VV-mode (1st argument as bias channel and 2nd argument as sense channel).
            
        
        Returns
        -------
		None
        """
        default_channels = {0: (1, 2), 1: (1,), 2: (1,)}
        if mode in [0, 1, 2]:
            if len(channels) == int(not mode)+1:
                logging.debug('{!s}: Set sweep channels to {!s}'.format(__name__, channels))
                self._sweep_channels = channels
            elif len(channels) == 0:
                logging.debug('{!s}: Set sweep channels to {!s}'.format(__name__, default_channels[mode]))
                self._sweep_channels = default_channels[mode]
            else:
                logging.error('{!s}: Cannot set sweep channels to {!s}'.format(__name__, channels))
                raise ValueError('{!s}: Cannot set sweep channels to {!s}'.format(__name__, channels))
            logging.debug('{!s}: Set sweep mode to {:d}'.format(__name__, mode))
            self._sweep_mode = mode
            self.__set_defaults_docstring()
        else:
            logging.error('{!s}: Cannot set sweep mode to {!s}'.format(__name__, mode))
            raise ValueError('{!s}: Cannot set sweep mode to {!s}'.format(__name__, mode))
        return
    
    def get_sweep_mode(self):
        """
        Gets an internal variable <mode> to decide weather
         * voltage is both applied and measured (VV-mode),
         * current is applied and voltage is measured (IV-mode) or
         * voltage is applied and current is measured (VI-mode).
        
        Parameters
        ----------
    	None
        
        Returns
        -------
        mode: int
            Sweep mode denoting bias and sense modes. Meanings are 0 (VV-mode), 1 (IV-mode) or 2 (VI-mode).
        """
        return self._sweep_mode
    
    def get_sweep_channels(self):
        """
        Gets channels <channels> that are used during sweeps. In the IV and VI sweep mode one channel requires for biasing and sensing, whereas two channels are returned in the VV sweep mode (1st as bias and 2nd as sense channel).
        
        Parameters
        ----------
		None
        
        Returns
        -------
        channels: int or tuple(int)
            Number of channel of usage. Meanings are 1 or 2 for IV-mode and VI-mode or a tuple containing both channels for VV-mode (1st argument as bias channel and 2nd argument as sense channel).
        """
        return self._sweep_channels
    
    def get_sweep_bias(self):
        """
        Calls get_bias_mode of channel <channel>. This method is needed for qkit.measure.transport.transport.py in case of no virtual tunnel electronic.
        
        Parameters
        ----------
		None
        
        Returns
        -------
        mode: int
            Bias mode. Meanings are 0 (current) and 1 (voltage).
        """
        return self.get_bias_mode(self._sweep_channels[0])
    
    def _set_sweep_start(self, val, channel=1):
        """
        Sets sweep start value of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Start value of sweep.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce[:VOLTage]:SWEep:STARt <voltage>|MINiumum|MAXimum
        try:
            logging.debug('{!s}: Set sweep start of channel {:d} to {:f}'.format(__name__, channel, val))
            # self._write(':chan{:d}:sour:swe:star {:f}'.format(channel, val))
            self._write(':chan{:d}:sour:{:s}:swe:star {:f}'.format(channel, self._IV_modes[self.get_bias_mode(channel)], val))
        except Exception as e:
            logging.error('{!s}: Cannot set sweep start of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set sweep start of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def _get_sweep_start(self, channel=1):
        """
        Gets sweep start value of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Start value of sweep.
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce[:VOLTage]:SWEep:STARt <voltage>|MINiumum|MAXimum
        try:
            logging.debug('{!s}: Get sweep start of channel {:d}'.format(__name__, channel))
            # return float(self._ask(':chan{:d}:sour:swe:star'.format(channel)))
            return float(self._ask(':chan{:d}:sour:{:s}:swe:star'.format(channel, self._IV_modes[self.get_bias_mode(channel)])))
        except Exception as e:
            logging.error('{!s}: Cannot get sweep start of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sweep start of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def _set_sweep_stop(self, val, channel=1):
        """
        Sets sweep stop value of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Stop value of sweep.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce[:VOLTage]:SWEep:STARt <voltage>|MINiumum|MAXimum
        try:
            logging.debug('{!s}: Set sweep stop of channel {:d} to {:f}'.format(__name__, channel, val))
            # self._write(':chan{:d}:sour:swe:stop {:f}'.format(channel, val))
            self._write(':chan{:d}:sour:{:s}:swe:stop {:f}'.format(channel, self._IV_modes[self.get_bias_mode(channel)], val))
        except Exception as e:
            logging.error('{!s}: Cannot set sweep stop of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set sweep stop of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def _get_sweep_stop(self, channel=1):
        """
        Gets sweep stop value of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: float
            Stop value of sweep.
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce[:VOLTage]:SWEep:STARt <voltage>|MINiumum|MAXimum
        try:
            logging.debug('{!s}: Get sweep stop of channel {:d}'.format(__name__, channel))
            # return float(self._ask(':chan{:d}:sour:swe:stop'.format(channel)))
            return float(self._ask(':chan{:d}:sour:{:s}:swe:stop'.format(channel, self._IV_modes[self.get_bias_mode(channel)])))
        except Exception as e:
            logging.error('{!s}: Cannot get sweep stop of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sweep stop of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def _set_sweep_step(self, val, channel=1):
        """
        Sets sweep step value of channel <channel> to <val>.
        
        Parameters
        ----------
        val: float
            Step value of sweep.
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce[:VOLTage]:SWEep:STARt <voltage>|MINiumum|MAXimum
        try:
            logging.debug('{!s}: Set sweep step of channel {:d} to {:f}'.format(__name__, channel, val))
            self._write(':chan{:d}:sour:{:s}:swe:step {:f}'.format(channel, self._IV_modes[self.get_bias_mode(channel)], val))
            # self._write(':chan{:d}:sour:swe:step {:f}'.format(channel, val))
        except Exception as e:
            logging.error('{!s}: Cannot set sweep step of channel {!s} to {!s}'.format(__name__, channel, val))
            raise type(e)('{!s}: Cannot set sweep step of channel {!s} to {!s}\n{!s}'.format(__name__, channel, val, e))
        return
    
    def _get_sweep_step(self, channel=1):
        """
        Gets sweep step value of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val:
            Step value of sweep.
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce[:VOLTage]:SWEep:STARt <voltage>|MINiumum|MAXimum
        try:
            logging.debug('{!s}: Get sweep step of channel {:d}'.format(__name__, channel))
            return float(self._ask(':chan{:d}:sour:{:s}:swe:step'.format(channel, self._IV_modes[self.get_bias_mode(channel)])))
            # return float(self._ask(':chan{:d}:sour:swe:step'.format(channel)))
        except Exception as e:
            logging.error('{!s}: Cannot get sweep step of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sweep step of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def _get_sweep_nop(self, channel=1):
        """
        Gets sweeps number of points (nop) of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val:
            Number of points of sweep.
        """
        try:
            logging.debug('{!s}: Get sweep nop of channel {:d}'.format(__name__, channel))
            return int((self._get_sweep_stop(channel=channel) - self._get_sweep_start(channel=channel)) / self._get_sweep_step(channel=channel) + 1)
        except Exception as e:
            logging.error('{!s}: Cannot get sweep nop of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sweep nop of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def set_sweep_parameters(self, sweep):
        """
        Sets sweep parameters <sweep> and prepares instrument for the set sweep mode.
        
        Parameters
        ----------
        sweep: array_likes of floats
            Sweep range containing start, stop and step size (e.g. sweep object using qkit.measure.transport.transport.sweep class)
        
        Returns
        -------
		None
        """
        # Corresponding Command: [:CHANnel<n>]:SOURce:MODE FIXed|SWEep|LIST|SINGle
        # Corresponding Command: [:CHANnel<n>]:SWEep:TRIGger EXTernal|AUXiliary|TIMer1|TIMer2|SENSe
        # Corresponding Command: :TRACe:POINts <integer>|MINimum|MAXimum
        # Corresponding Command: :TRACe:CHANnel<n>:DATA:FORMat ASCii|BINary
        # Corresponding Command: :TRACe:BINary:REPLy BINary|ASCii
        # Corresponding Command: :TRACe[:STATe] 1|0|ON|OFF
        if not self._sweep_mode:  # 0 (VV-mode)
            channel_bias, channel_sense = self._sweep_channels
            self.set_sync(val=True)
        elif self._sweep_mode in [1, 2]:  # 1 (IV-mode) | 2 (VI-mode)
            channel_bias, channel_sense = self._sweep_channels*2
        try:
            logging.debug('{!s}: Set sweep parameters of channel {:d} and {:d} to {!s}'.format(__name__, channel_bias, channel_sense, sweep))
            self._set_sweep_start(val=float(sweep[0]), channel=channel_bias)
            self._set_sweep_stop(val=float(sweep[1]), channel=channel_bias)
            self._set_sweep_step(val=np.abs(float(sweep[2])), channel=channel_bias)
            self.set_bias_trigger(mode='sens', channel=channel_bias)
            self.set_sense_trigger(mode='sour', channel=channel_sense)
            self.set_bias_value(val=self._get_sweep_start(channel=channel_bias), channel=channel_bias)
            self._write(':chan{:d}:sour:mode swe'.format(channel_bias))
            self._write(':chan{:d}:swe:trig ext'.format(channel_bias))
            self._write(':trac:poin max')  # alternative: self._write(':trac:poin {:d}'.format(self._get_sweep_nop(channel=channel_bias)))
            self._write(':trac:chan{:d}:data:form asc'.format(channel_sense))
            self._write(':trac:bin:repl asc')
        except Exception as e:
            logging.error('{!s}: Cannot set sweep parameters of channel {:d} and {:d} to {!s}'.format(__name__, channel_bias, channel_sense, sweep))
            raise type(e)('{!s}: Cannot set sweep parameters of channel {:d} and {:d} to {!s}\n{!s}'.format(__name__, channel_bias, channel_sense, sweep, e))
        return
    
    def get_tracedata(self):
        """
        Starts bias sweep with already set parameters and gets trace data of bias <bias_values> and sense <sense_values> in the set sweep mode.
        
        Parameters
        ----------
		None
        
        Returns
        -------
        bias_values: numpy.array(float)
            Measured bias values.
        sense_values: numpy.array(float)
            Measured sense values.
        """
        # Corresponding Command: [:CHANnel<n>]:INITiate [DUAL]
        # Corresponding Command: :STARt
        # Corresponding Command: :TRACe[:STATe] 1|0|ON|OFF
        # Corresponding Command: :TRACe:CHANnel<n>:DATA:READ? [TM|DO|DI|SF|SL|MF|ML|LC|HC|CP]
        if not self._sweep_mode:  # 0 (VV-mode)
            channel_bias, channel_sense = self._sweep_channels
        elif self._sweep_mode in [1, 2]:  # 1 (IV-mode) | 2 (VI-mode)
            channel_bias, channel_sense = self._sweep_channels*2
        try:
            logging.debug('{!s}: Take sweep data of channel {:d} and {:d}'.format(__name__, channel_bias, channel_sense))
            self._write(':chan{:d}:init'.format(channel_bias))
            self._wait_for_ready_for_sweep(channel=channel_bias)
            self._write(':trac:stat 1')
            self._wait_for_OPC()
            time.sleep(100e-6)
            self._write(':star')
            self._wait_for_end_of_sweep(channel=channel_bias)
            time.sleep(self.get_sense_delay(channel=channel_sense))
            self._wait_for_end_of_measure(channel=channel_sense)
            self._write(':trac:stat 0')
            bias_values = np.fromstring(string=self._ask('trac:chan{:d}:data:read? sl'.format(channel_bias)), dtype=float, sep=',')
            sense_values = np.fromstring(string=self._ask('trac:chan{:d}:data:read? ml'.format(channel_sense)), dtype=float, sep=',')
            return bias_values, sense_values
        except Exception as e:
            logging.error('{!s}: Cannot take sweep data of channel {!s} and {!s}'.format(__name__, channel_bias, channel_sense))
            raise type(e)('{!s}: Cannot take sweep data of channel {!s} and {!s}\n{!s}'.format(__name__, channel_bias, channel_sense, e))
    
    def take_IV(self, sweep):
        """
        Takes IV curve with sweep parameters <sweep> in set sweep mode.
        
        Parameters
        ----------
        sweep: array_likes of floats
            Sweep range containing start, stop and step size (e.g. sweep object using qkit.measure.transport.transport.sweep class)
        
        Returns
        -------
        bias_values: numpy.array(float)
            Measured bias values.
        sense_values: numpy.array(float)
            Measured sense values.
        """
        self.set_sweep_parameters(sweep=sweep)
        return self.get_tracedata()
    
    def get_bias_ccr(self):
        """
        Gets the entire bias condition code register (ccr)
        
        Parameters
        ----------
        None
        
        Returns
        -------
        ccr: list of booleans
            Condition code register of bias. Entries are
            0:  CH1 End of Sweep
            1:  CH1 Ready for Sweep
            2:  CH1 Low Limiting
            3:  CH1 High Limiting
            4:  CH1 Tripped
            5:  CH1 Emergency (Temperature/Current over)
            6:  ---
            7:  ---
            8:  CH2 End of Sweep
            9:  CH2 Ready for Sweep
            10: CH2 Low Limiting
            11: CH2 High Limiting
            12: CH2 Tripped
            13: CH2 Emergency (Temperature/Current over)
            14: Inter Locking
            15: Start Sampling Error
        """
        # Corresponding Command: :STATus:SOURce:CONDition?
        try:
            logging.debug('{!s}: Get bias ccr'.format(__name__))
            ccr = int(self._ask(':stat:sour:cond'))
            ans = []
            for i in range(16):
                ans.append(bool((ccr >> i) % 2))
            return ans
        except Exception as e:
            logging.error('{!s}: Cannot get bias ccr'.format(__name__))
            raise type(e)('{!s}: Cannot get bias ccr\n{!s}'.format(__name__, e))
    
    def print_bias_ccr(self):
        """
        Prints the entire bias condition code register (ccr) including explanation.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        ccr = self.get_bias_ccr()
        msg = [('\n\t{:s}:\t{!r}\t({:s})'.format(sb[0], ccr[i], sb[1])) for i, sb in enumerate(self.bias_ccr) if sb != ('', '')]
        print('Bias ccr :{:s}'.format(''.join(msg)))
        return
    
    def get_sense_ccr(self):
        """
        Gets the entire sense condition code register (ccr)
        
        Parameters
        ----------
        None
        
        Returns
        -------
        ccr: list of booleans
            Condition code register of sense. Entries are
                0:  CH1 End of Measure
                1:  ---
                2:  CH1 Compare result is Low
                3:  CH1 Compare result is High
                4:  ---
                5:  CH1 Over Range
                6:  ---
                7:  ---
                8:  CH2 End of Measure
                9:  ---
                10: CH2 Compare result is Low
                11: CH2 Compare result is High
                12: ---
                13: CH2 Over Range
                14: End of Trace
                15: Trigger Sampling Error
        """
        # Corresponding Command: :STATus:SENSe:CONDition?
        try:
            logging.debug('{!s}: Get sense ccr'.format(__name__))
            ccr = int(self._ask(':stat:sens:cond'))
            ans = []
            for i in range(16):
                ans.append(bool((ccr >> i) % 2))
            return ans
        except Exception as e:
            logging.error('{!s}: Cannot get sense ccr'.format(__name__))
            raise type(e)('{!s}: Cannot get sense ccr\n{!s}'.format(__name__, e))
    
    def print_sense_ccr(self):
        """
        Prints the entire bias condition code register (ccr) including explanation.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        ssr = self.get_sense_ccr()
        msg = [('\n\t{:s}:\t{!r}\t({:s})'.format(sb[0], ssr[i], sb[1])) for i, sb in enumerate(self.sense_ccr) if sb != ('', '')]
        print('Sense ccr:{:s}'.format(''.join(msg)))
        return
    
    def get_end_of_sweep(self, channel=1):
        """
        Gets event of bias condition code register (ccr) entry "End for Sweep" <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: bool
            End of sweep entry of bias condition code register.
        """
        # Corresponding Command: :STATus:SOURce:EVENt?
        try:
            logging.debug('''{!s}: Get bias ccr event "End of Sweep" of channel {:d}'''.format(__name__, channel))
            return bool((int(self._ask(':stat:sour:even')) >> (0+8*(channel-1))) % 2)
        except Exception as e:
            logging.error('{!s}: Cannot get bias ccr event "End of Sweep" of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get bias ccr event "End of Sweep" of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def _wait_for_end_of_sweep(self, channel=1):
        """
        Waits until the event of bias condition code register (ccr) entry "End for Sweep" of channel <channel> occurs.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        while not (self.get_end_of_sweep(channel=channel)):
            time.sleep(100e-3)
        return
    
    def get_ready_for_sweep(self, channel=1):
        """
        Gets condition of bias condition code register (ccr) entry "Ready for Sweep" <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: bool
            End of sweep entry of bias condition code register.
        """
        # Corresponding Command: :STATus:SOURce:CONDition?
        try:
            logging.debug('''{!s}: Get bias ccr event "Ready of Sweep" of channel {:d}'''.format(__name__, channel))
            return bool((int(self._ask(':stat:sour:cond')) >> (1+8*(channel-1))) % 2)
        except Exception as e:
            logging.error('{!s}: Cannot get bias ccr condition "Ready for Sweep" of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get bias ccr condition "Ready for Sweep" of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def _wait_for_ready_for_sweep(self, channel=1):
        """
        Waits until the condition of bias condition code register (ccr) entry "Ready for Sweep" of channel <channel> is fullfilled.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        while not (self.get_ready_for_sweep(channel=channel)):
            time.sleep(100e-3)
        return
    
    def get_end_of_measure(self, channel=1):
        """
        Gets condition of sense condition code register (ccr) entry "End of Measure" <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: bool
            End of sweep entry of bias condition code register.
        """
        # Corresponding Command: :STATus:SENSe:CONDition?
        try:
            logging.debug('''{!s}: Get sense ccr event "End of Measure" of channel {:d}'''.format(__name__, channel))
            return bool((int(self._ask(':stat:sens:cond')) >> (0+8*(channel-1))) % 2)
        except Exception as e:
            logging.error('{!s}: Cannot get sense ccr event "End of Measure" of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sense ccr event "End of Measure" of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def _wait_for_end_of_measure(self, channel=1):
        """
        Waits until the condition of sense condition code register (ccr) entry "End for Measure" of channel <channel> is fullfilled.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        while not (self.get_end_of_measure(channel=channel)):
            time.sleep(100e-3)
        return
    
    def get_end_of_trace(self, channel=1):
        """
        Gets condition of sense condition code register (ccr) entry "End of Trace" <val> of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        val: bool
            End of sweep entry of bias condition code register.
        """
        # Corresponding Command: :STATus:SENSe:CONDition?
        try:
            logging.debug('''{!s}: Get sense ccr event "End of Trace" of channel {:d}'''.format(__name__, channel))
            return bool((int(self._ask(':stat:sens:cond')) >> 14) % 2)
        except Exception as e:
            logging.error('{!s}: Cannot get sense ccr event "End of Trace"  of channel {!s}'.format(__name__, channel))
            raise type(e)('{!s}: Cannot get sense ccr event "End of Trace"  of channel {!s}\n{!s}'.format(__name__, channel, e))
    
    def _wait_for_end_of_trace(self, channel=1):
        """
        Waits until the condition of sense condition code register (ccr) entry "End for Trace" of channel <channel> is fullfilled.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
        None
        """
        while not (self.get_end_of_trace(channel=channel)):
            time.sleep(100e-3)
        return
    
    def get_OPC(self):
        """
        Waits until the condition of standard event register entry "Operation complete" is fulfilled.
        
        Parameters
        ----------
		None
        
        Returns
        -------
		None
        """
        # Corresponding Command: *OPC
        try:
            logging.debug('''{!s}: Get ccr event "Operation complete"'''.format(__name__))
            return bool(int(self._ask('*OPC')))
        except Exception as e:
            logging.error('{!s}: Cannot get ccr condition "Operation complete"'.format(__name__))
            raise type(e)('{!s}: Cannot get ccr condition "Operation complete"\n{!s}'.format(__name__, e))
    
    def _wait_for_OPC(self):
        """
        Waits until the condition of standard event register entry "Operation complete" is fulfilled.
        
        Parameters
        ----------
		None
        
        Returns
        -------
		None
        """
        while not (self.get_OPC()):
            time.sleep(1e-3)
        return
    
    def set_defaults(self, sweep_mode=None):
        """
        Sets default settings. Actual settings are:
        default parameters
        
        Parameters
        ----------
        mode: int
            Sweep mode denoting bias and sense modes. Possible values are 0 (VV-mode), 1 (IV-mode) or 2 (VI-mode). Default is 1
        
        Returns
        -------
		None
        """
        self.reset()
        # beeper off
        self._write(':syst:beep 0')
        # distiguish different sweep modes
        if sweep_mode is not None:
            self.set_sweep_mode(sweep_mode)
        # set values
        for i, channel in enumerate(self._sweep_channels):
            for func, param in self._defaults[self._sweep_mode][i].items():
                func(param, channel=channel)
        return
    
    def __set_defaults_docstring(self):
        '''
        Sets docstring of "set_defaults" method with actual default settings.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        '''
        new = 'sweep mode: {:d}:\n'.format(self._sweep_mode) + ''.join(['\t    channel: {:d}\n'.format(channel) + ''.join(['\t\t{:s}: {!s}\n'.format(key_parameter, val_parameter) for key_parameter, val_parameter in self._defaults[self._sweep_mode][i].items()]) for i, channel in enumerate(self._sweep_channels)])
        self.set_defaults.__func__.__doc__ = self.set_defaults.__func__.__doc__.replace(self._default_str, new)
        self._default_str = new
        return
    
    def get_all(self, channel=1):
        """
        Prints all settings of channel <channel>.
        
        Parameters
        ----------
        channel: int
            Number of channel of interest. Must be 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        logging.debug('{:s}: Get all'.format(__name__))
        print('synchronization    = {!r}'.format(self.get_sync()))
        print('measurement mode   = {:s}'.format(self._measurement_modes[self.get_measurement_mode(channel=channel)]))
        print('bias mode          = {:s}'.format(self._IV_modes[self.get_bias_mode(channel=channel)]))
        print('sense mode         = {:s}'.format(self._IV_modes[self.get_sense_mode(channel=channel)]))
        print('bias range         = {:1.0e}{:s}'.format(self.get_bias_range(channel=channel), self._IV_units[self.get_bias_mode(channel=channel)]))
        print('sense range        = {:1.0e}{:s}'.format(self.get_sense_range(channel=channel), self._IV_units[self.get_sense_mode(channel=channel)]))
        print('bias delay         = {:1.3e}s'.format(self.get_bias_delay(channel=channel)))
        print('sense delay        = {:1.3e}s'.format(self.get_sense_delay(channel=channel)))
        print('sense average      = {:d}'.format(self.get_sense_average(channel=channel)[1]))
        print('plc                = {:f}Hz'.format(self.get_plc()))
        print('sense nplc         = {:f}'.format(self.get_sense_nplc(channel=channel)))
        print('sense autozero     = {!r}'.format(self.get_sense_autozero(channel=channel)))
        print('status             = {!r}'.format(self.get_status(channel=channel)))
        if self.get_status(channel=channel):
            print('bias value         = {:f}{:s}'.format(self.get_bias_value(channel=channel), self._IV_units[self.get_bias_mode(channel=channel)]))
            print('sense value        = {:f}{:s}'.format(self.get_sense_value(channel=channel), self._IV_units[self.get_sense_mode(channel=channel)]))
        print('sweep start        = {:f}{:s}'.format(self._get_sweep_start(channel=channel), self._IV_units[self.get_bias_mode(channel=channel)]))
        print('sweep stop         = {:f}{:s}'.format(self._get_sweep_stop(channel=channel), self._IV_units[self.get_bias_mode(channel=channel)]))
        print('sweep step         = {:f}{:s}'.format(self._get_sweep_step(channel=channel), self._IV_units[self.get_bias_mode(channel=channel)]))
        print('sweep nop          = {:d}'.format(self._get_sweep_nop(channel=channel)))
        for err in self.get_error():
            print('error              = {:d}\t{:s}'.format(err[0], err[1]))
        self.print_bias_ccr()
        self.print_sense_ccr()
        return
    
    def reset(self):
        """
        Resets the instrument or a single channel to default conditions.
        
        Parameters
        ----------
        channel: int
            Number of channel to be set. Must be None, 1 or 2. Default is 1.
        
        Returns
        -------
		None
        """
        # Corresponding Command: *RST
        try:
            logging.debug('{!s}: Reset instrument to factory settings'.format(__name__))
            self._write('*RST')
        except Exception as e:
            logging.error('{!s}: Cannot reset instrument to factory settings'.format(__name__))
            raise type(e)('{!s}: Cannot reset instrument to factory settings\n{!s}'.format(__name__, e))
        return
    
    def get_error(self):
        """
        Gets errors of instrument and removes them from the error queue.
        
        Parameters
        ----------
		None
        
        Returns
        -------
        err: array of str
            Entries from the error queue
        """
        # Corresponding Command: :SYSTem:ERRor?
        try:
            logging.debug('{!s}: Get errors of instrument'.format(__name__))
            err = [self._visainstrument.query(':syst:err?').rstrip().split(',', 1)]
            while err[-1] != ['0', '"No error"']:
                err.append(self._visainstrument.query(':syst:err?').rstrip().split(',', 1))
            if len(err) > 1:
                err = err[:-1]
            err = [[int(e[0]), str(e[1][1:-1])] for e in err]
            return err
        except Exception as e:
            logging.error('{!s}: Cannot get errors of instrument'.format(__name__))
            raise type(e)('{!s}: Cannot get errors of instrument\n{!s}'.format(__name__, e))
    
    def _raise_error(self):
        """
        Gets errors of instrument and as the case may be raises a python error.
        
        Parameters
        ----------
		None
        
        Returns
        -------
		None
        """
        errors = self.get_error()
        if len(errors) == 1 and errors[0][0] is 0:  # no error
            return
        else:
            msg = __name__ + ' raises the following errors:'
            for err in errors:
                msg += '\n{:s} ({:s})'.format(self.err_msg[err[0]], err[1])
            raise ValueError(msg)
    
    def clear_error(self):
        """
        Clears error of instrument.
        
        Parameters
        ----------
		None
        
        Returns
        -------
		None
        """
        # Corresponding Command: *CLS
        try:
            logging.debug('{!s}: Clears error of instrument'.format(__name__))
            self._write('*CLS')
        except Exception as e:
            logging.error('{!s}: Cannot clears error of instrument'.format(__name__))
            raise type(e)('{!s}: Cannot clears error of instrument\n{!s}'.format(__name__, e))
        return
    
    def close(self):
        """
        Closes the VISA-instrument to disconnect the instrument.
        
        Parameters
        ----------
		None
        
        Returns
        -------
		None
        """
        try:
            logging.debug('{!s}: Close VISA-instrument'.format(__name__))
            self._visainstrument.close()
        except Exception as e:
            logging.error('{!s}: Cannot close VISA-instrument'.format(__name__))
            raise type(e)('{!s}: Cannot close VISA-instrument\n{!s}'.format(__name__, e))
        return
    
    def get_parameters(self):
        """
        Gets a parameter list <parlist> of measurement specific setting parameters.
        Needed for .set-file in 'write_additional_files', if qt parameters are not used.
        
        Parameters
        ----------
		None
        
        Returns
        -------
        parlist: dict
            Parameter names as keys, corresponding channels of interest as values.
        """
        parlist = {'measurement_mode': [1, 2],
                   'sync': {'channels': [None]},
                   'bias_mode': {'channels': [1, 2]},
                   'sense_mode': {'channels': [1, 2]},
                   'bias_range': {'channels': [1, 2]},
                   'sense_range': {'channels': [1, 2]},
                   'bias_trigger': {'channels': [1, 2]},
                   'sense_trigger': {'channels': [1, 2]},
                   'bias_delay': {'channels': [1, 2]},
                   'sense_delay': {'channels': [1, 2]},
                   'sense_average': {'channels': [1, 2]},
                   'bias_value': {'channels': [1, 2]},
                   'plc': {'channels': [None]},
                   'sense_nplc': {'channels': [1, 2]},
                   'sense_autozero': {'channels': [1, 2]},
                   'status': {'channels': [1, 2]}}
        return parlist
    
    def get(self, param, **kwargs):
        """
        Gets the current parameter <param> by evaluation 'get_'+<param> and corresponding channel if needed
        In combination with <self.get_parameters> above.
        
        Parameters
        ----------
        param: str
            Parameter of interest.
        channels: array_likes
            Number of channels of interest. Must be in {1, 2} for channel specific parameters or None for channel independant (global) parameters.
        
        Returns
        -------
        parlist: dict
            Parameter names as keys, values of corresponding channels as values.
        """
        # ugly fix for kwargs sometimes having format {"channels": *ints*, ...} and other times {"channels": {"channels": *ints*}, ...} 
        channels = kwargs.get('channels')
        if type(channels) == dict:
            channels = channels.get('channels')
        if channels == [None]:
            return getattr(self, 'get_{!s}'.format(param))()
        else:
            return tuple([getattr(self, 'get_{!s}'.format(param))(channel=channel) for channel in channels])
    



# This instrument transmits and receives raw ethernet packets to and from 
# a network interface. It is used to communicate with GHZdac devices.
# Markus Jerger <jerger@kit.edu>, 2011
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# TO DO
# - cleanup/timeout of registered listeners
#
#

from qkit.core.instrument_base import Instrument
import types
import logging
# comms and byte packing
import threading
import socket
import dpkt
import struct
# ip filter
import re

class EthernetServer(Instrument):
    '''
    This instrument passes packets between UDP and Ethernet, to talk to GHZdac devices.

    Usage:
    Initialize with
    <name> = qt.instruments.create('name', 'EthernetServer', interface='<interface name>', ethertype=<ethertype>, ip_host=<ip_host>, ip_port=<ip_port>)
    <interface name> = name of the raw packet network interface
    <ethertype> = 0x0003 (ETH_P_ALL, receive all packets) or any other EtherType value
    <ip_host>:<ip_port> = 127.0.0.1:413 to bind the UDP interface to
    '''

    # broadcast address
    MAC_BROADCAST = struct.pack('>6B', *[0xff, 0xff, 0xff, 0xff, 0xff, 0xff])
    # capture all packets ethertype
    ETH_P_ALL = 0x0003
    # maximum packet size
    RECV_SIZE = 2048

    def __init__(self, name, interface, ethertype = 0x0003, ip_host = '127.0.0.1', ip_port = 413, ip_filter = ''):
        '''
        Initializes the UDP-to-Ethernet bridge

        Input:
            name (string)    : name of the instrument
            interface (string) : raw ethernet network interface
            ethertype : ethernet frame type to receive
            ip_host:ip_port : UDP interface address
            ip_filter : regular expression used to match incoming packets against

        Output:
            None
        '''
        logging.debug(__name__ + ' : Initializing instrument')
        Instrument.__init__(self, name, tags=['physical'])

        self._interface = interface
        self._ethertype = int(ethertype)
        self._ip_host = ip_host
        self._ip_port = int(ip_port)
        self.do_set_ip_filter(ip_filter)
        self._clients = {} # associate MAC addresses with udp clients
        self._devices = set() # mac addresses seen on the raw interface
        self._updateLock = threading.Lock()

        # add parameters
        self.add_parameter('interface', type=str, flags=Instrument.FLAG_GET)
        self.add_parameter('MAC', type=str, flags=Instrument.FLAG_GET)
        self.add_parameter('ethertype', type=int, flags=Instrument.FLAG_GET)
        self.add_parameter('ip_host', type=str, flags=Instrument.FLAG_GET)
        self.add_parameter('ip_port', type=int, flags=Instrument.FLAG_GET)
        self.add_parameter('ip_filter', type=str, flags=Instrument.FLAG_GETSET)
        self.add_parameter('devices', type=types.ListType, flags=Instrument.FLAG_GET)
        self.add_parameter('listeners', type=types.ListType, flags=Instrument.FLAG_GET)
        self.add_parameter('raw_rcvbuf', type=int, flags=Instrument.FLAG_GETSET)
        self.add_parameter('udp_rcvbuf', type=int, flags=Instrument.FLAG_GETSET)
        self.add_function('get_all')

        # open sockets
        self._init_udp()
        self._init_raw()

        # update ui
        self.get_all()

        # start daemons
        self._udp_receiver = self.udp_receiver(self)
        self._udp_receiver.daemon = True
        self._udp_receiver.start()
        self._raw_receiver = self.raw_receiver(self)
        self._raw_receiver.daemon = True
        self._raw_receiver.start()

    def __del__(self):
        '''
        clean up
        '''
        self._udp_socket.close()
        self._raw_socket.close()

    class udp_receiver(threading.Thread):
        '''
        receive client requests
        '''
        def __init__(self, node):
            threading.Thread.__init__(self, name = 'node_udp')
            self.node = node

        def run(self):
            logging.debug(self.name + ' : udp listener started.')
            while(True):
                # receive packet (ethernet packet in the payload of an UDP packet)
                (raw_packet, sender) = self.node._udp_socket.recvfrom(self.node.RECV_SIZE)
                packet = dpkt.ethernet.Ethernet(raw_packet)
                logging.debug(self.name + ' : udp packet received.')
                # source filtering
                if(not self.node._ip_filter(sender)):
                    logging.debug(self.name + ' : udp packet from %s discarded.'%sender[0])
                    continue
                # add this connection to the listeners of a specific device MAC
                if(packet.dst not in self.node._clients):
                    self.node._updateLock.acquire(True)
                    if(packet.dst not in self.node._clients):
                        self.node._clients[packet.dst] = set()
                    self.node._updateLock.release()
                if(sender not in self.node._clients[packet.dst]):
                    self.node._clients[packet.dst].add(sender)
                # dispatch to raw socket
                packet.src = self.node._MAC
                self.node._raw_socket.send(packet.pack())

    class raw_receiver(threading.Thread):
        '''
        receive raw network packets
        '''
        def __init__(self, node):
            threading.Thread.__init__(self, name = 'node_raw')
            self.node = node

        def run(self):
            logging.debug(self.name + ' : raw listener started.')
            while(True):
                # receive packet (ethernet packet)
                (raw_packet, sender) = self.node._raw_socket.recvfrom(self.node.RECV_SIZE)
                packet = dpkt.ethernet.Ethernet(raw_packet)
                logging.debug(self.name + ' : raw packet received.')
                self.node._devices.add(packet.src)
                # dispatch to all listeners
                for client_list in [self.node._clients.get(packet.src, None), self.node._clients.get(self.node.MAC_BROADCAST, None)]:
                    if(client_list == None): continue
                    for client in client_list:
                        self.node._udp_socket.sendto(raw_packet, client)

    def _init_udp(self):
        # open udp socket
        logging.debug(__name__ + ' : Listening for packets on %s:%d.'%(self._ip_host, self._ip_port))
        self._udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._udp_socket.bind((self._ip_host, self._ip_port))

    def _init_raw(self):
        # open raw socket
        if(self._ethertype == self.ETH_P_ALL):
            logging.info(__name__ + ' : EtherType set to ETH_P_ALL, this may generate high CPU load')
	    self._raw_socket = socket.socket(socket.PF_PACKET, socket.SOCK_RAW, self._ethertype)
        self._raw_socket.bind((self._interface, self._ethertype))
        self._MAC = self._raw_socket.getsockname()[4]

    def get_all(self):
        ''' call all getters to update the ui '''
        self.get_interface()
        self.get_MAC()
        self.get_ethertype()
        self.get_ip_host()
        self.get_ip_port()
        self.get_ip_filter()
        self.get_udp_rcvbuf()
        self.get_raw_rcvbuf()

        self.get_listeners()
        self.get_devices()

    def do_get_interface(self):
        ''' return name of the network interface used '''
        return self._interface

    def do_get_MAC(self):
        ''' return own MAC address '''
        return '%02x:%02x:%02x:%02x:%02x:%02x'%(struct.unpack('>6B', self._MAC))
 
    def do_get_ethertype(self):
        ''' return packet type received '''
        return self._ethertype

    def do_get_ip_host(self):
        ''' return hostname of the udp socket '''
        return self._ip_host

    def do_get_ip_port(self):
        ''' return port number of the udp socket '''
        return self._ip_port

    def do_set_ip_filter(self, expr):
        ''' udp source filter (regular) expression '''
        if(expr):
            self._ip_filter_expr = expr
            self._ip_filter = re.compile(expr).match
        else:
            self._ip_filter_expr = ''
            self._ip_filter = lambda s: True

    def do_get_ip_filter(self):
        ''' return udp source filter expression '''
        return self._ip_filter_expr

    def do_get_listeners(self):
        ''' return list of udp clients '''
        # gather list of clients in <ip>:<port> format
        listeners = []
        for client_list in self._clients.values():
            listeners.append(['%s:%d'%(x[0], x[1]) for x in list(client_list)])
        # remove duplicates
        return list(set(listeners))

    def do_get_devices(self):
        ''' return list of seen devices '''
        #return self._clients.keys()
        return list(self._devices)

    def do_set_raw_rcvbuf(self, bufsize):
        ''' set the receive buffer size of the raw socket '''
        self._raw_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, int(bufsize))

    def do_get_raw_rcvbuf(self):
        ''' get the receive buffer size of the raw socket '''
        return self._raw_socket.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)

    def do_set_udp_rcvbuf(self, bufsize):
        ''' set the receive buffer size of the raw socket '''
        self._udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, int(bufsize))

    def do_get_udp_rcvbuf(self):
        ''' get the receive buffer size of the raw socket '''
        return self._udp_socket.getsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF)




# -*- coding: utf-8 -*-
# spectroscopy.py analysis class for qkit spectroscopy measurement data
# Micha Wildermuth, micha.wildermuth@kit.edu 2023

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import numpy as np
from qkit.analysis.qdata import qData
from qkit.analysis.circle_fit.circle_fit_2019 import circuit


class spectrum(qData):
    """
    This is an analysis class for spectrum-like spectroscopy measurements taken by
    `qkit.measure.spectroscopy.spectroscopy.py`.
    """

    def __init__(self):
        """
        Initializes an analysis class for spectrum-like spectroscopy measurements taken by
        `qkit.measure.spectroscopy.spectroscopy.py`.

        Parameters
        ----------
        None

        Returns
        -------
        None

        Examples
        --------
        >>> import numpy as np
        >>> import qkit
        QKIT configuration initialized -> available as qkit.cfg[...]
        >>> qkit.start()
        Starting QKIT framework ... -> qkit.core.startup
        Loading module ... S10_logging.py
        Loading module ... S12_lockfile.py
        Loading module ... S14_setup_directories.py
        Loading module ... S20_check_for_updates.py
        Loading module ... S25_info_service.py
        Loading module ... S30_qkit_start.py
        Loading module ... S65_load_RI_service.py
        Loading module ... S70_load_visa.py
        Loading module ... S80_load_file_service.py
        Loading module ... S85_init_measurement.py
        Loading module ... S98_started.py
        Loading module ... S99_init_user.py
        Initialized the file info database (qkit.fid) in 0.000 seconds.

        >>> from qkit.analysis.qdata import spectrum
        >>> s = spectrum()
        """
        super().__init__()
        self.circuit = circuit

    def load(self, uuid):
        """
        Loads qkit spectroscopy data with given uuid <uuid>.

        Parameters
        ----------
        uuid: str
            Qkit identification name, that is looked for and loaded.

        Returns
        -------
        None

        Examples
        --------
        >>> s.load(uuid='XXXXXX')
        """
        super().load(uuid=uuid)
        if self.m_type != 'spectroscopy':
            raise AttributeError('No spectroscopy data loaded. Use data acquired with spectroscopy measurement class or general qData class.')
        self.scan_dim = self.df.data.amplitude.attrs['ds_type']  # scan dimension (1D, 2D, ...)
        self._get_xy_parameter(self.df.data.amplitude)
        self.circlefit = None

    def open_qviewkit(self, uuid=None, ds=None):
        """
        Opens qkit measurement data with given uuid <uuid> in qviewkit.

        Parameters
        ----------
        uuid: str
            Qkit identification name, that is looked for and opened in qviewkit.
        ds: str | list(str)
            Datasets that are opened instantaneously. Default for spectroscopy data is 'amplitude' and 'phase'.

        Returns
        -------
        None
        """
        if ds is None:
            ds = [_ds if _ds in self.df.data.__dict__.keys() else None for _ds in ['amplitude', 'phase']]
        super().open_qviewkit(uuid=uuid, ds=ds)

    def setup_circlefit(self, type, f_data=None, z_data_raw=None):
        if f_data is None:
            f_data = self.frequency
        if z_data_raw is None:
            if hasattr(self, 'real') and hasattr(self, 'imag'):
                z_data_raw = self.real + 1j * self.imag
            elif hasattr(self, 'amplitude') and hasattr(self, 'phase'):
                z_data_raw = self.amplitude + np.exp(1j * self.phase)
            else:
                raise NameError('no S21 data available. Please load either real and imaginary data or amplitude and phase data.')
        self.circlefit = {'reflection': self.circuit.reflection_port,
                          'notch': self.circuit.notch_port}[type](f_data=f_data,
                                                                  z_data_raw=z_data_raw)





import qkit
from qkit.core.instrument_base import Instrument
from qkit import visa
import logging
import numpy as np
import Keysight_VNA_E5080B


class Keysight_VNA_E5080B_spectrum(Keysight_VNA_E5080B.Keysight_VNA_E5080B):
    def __init__(self, name, address, channel_index=1):
        super().__init__(name, address, channel_index)
        
        self.add_parameter('measurement_class', type=str,
            flags=Instrument.FLAG_GET)
        
        self.add_parameter('signal_format', type=int,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=1)
        self.signal_formats = {0: 'MLIN', 1: 'MLOG'}
        self.inv_signal_formats = {val.upper(): key for key, val in self.signal_formats.items()}
        
        self.add_parameter('signal_unit', type=int,
            flags=Instrument.FLAG_GETSET,
            minval=0, maxval=3)
        self.lin_units = {0: 'W', 1: 'V', 2: 'A'}
        self.inv_lin_units = {val.upper(): key for key, val in self.lin_units.items()}
        self.log_units =  {0: 'dBm', 1: 'dBmV', 2: 'dBuV', 3: 'dBmA'}
        self.inv_log_units = {val.upper(): key for key, val in self.log_units.items()}
        
        # Implement functions
        self.add_function('get_trace_name')
        self.add_function('get_y_unit')
        self.add_function('get_active_traces')
        self.add_function('get_tracedata')
        
    
    def write(self, cmd):
        self._visainstrument.write(cmd)
    
    def query(self, cmd):
        return self._visainstrument.query(cmd).strip('\n')
    
    def do_get_measurement_class(self):
        # corresponding command: SENSe<cnum>:CLASs:NAME?
        return self._visainstrument.query('SENS:CLAS:NAME?').strip().strip('"')
    
    def do_set_signal_format(self, signal_format):
        """
        Sets the signal format
        
        signal_format: int
            0: linear amplitude
            1: logarithmic amplitude
        """
        self.write('CALC:MEAS:FORM {:s}'.format(self.signal_formats[signal_format]))
    
    def do_get_signal_format(self):
        """
        Gets the signal format
        
        signal_format: int
            0: linear amplitude
            1: logarithmic amplitude
        """
        ans = self.query('CALC:MEAS:FORM?').strip()
        return self.inv_signal_formats[ans]
    
    def do_set_signal_unit(self, unit, signal_format=None):
        """
        Sets the signal unit
        
        unit: int
            for linear format
                0: Watt
                1: Volt
                2: Ampere
            for logarithmic format
                0: dBm
                1: dBmV
                2: dBuV
                3: dBmA
        """
        if signal_format is None:
            signal_format = self.get_signal_format()
        else:
            self.set_signal_format(signal_format)
        self.write('CALC:MEAS:FORM:UNIT {:s}, {:s}'.format(self.signal_formats[signal_format], {0: self.lin_units, 1: self.log_units}[signal_format][unit]))
            
    def do_get_signal_unit(self):
        """
        Gets the signal unit
        
        unit: int
            for linear format
                0: Watt
                1: Volt
                2: Ampere
            for logarithmic format
                0: dBm
                1: dBmV
                2: dBuV
                3: dBmA
        """
        signal_format = self.get_signal_format()
        ans = self.query('CALC:MEAS:FORM:UNIT? {:s}'.format(self.signal_formats[signal_format]))
        return {0: self.inv_lin_units, 1: self.inv_log_units}[signal_format][ans]
    
    def get_trace_name(self, trace=1):
        # corresponding command: CALCulate<cnum>:MEASure<mnum>:FORMat:UNIT <dataFormat>, <units>
        return 'amplitude'
    
    def get_y_unit(self, trace=1):
        # corresponding command: CALCulate<cnum>:MEASure<mnum>:FORMat:UNIT <dataFormat>, <units>
        signal_format = self.get_signal_format()
        unit = self.get_signal_unit()
        return {0: self.lin_units, 1: self.log_units}[signal_format][unit]
    
    def get_active_traces(self):
        # corresponding command: CALCulate<cnum>:PARameter:COUNt <value>
        return int(self.query('CALC:PAR:COUN?'))
        
    def get_tracedata(self, trace=1):
        if self.get_measurement_class() == 'Spectrum Analyzer':
            self._visainstrument.write('FORM:DATA REAL,32')
            self._visainstrument.write('FORM:BORD SWAPPED') #SWAPPED
            data = self._visainstrument.query_binary_values('CALC%i:MEAS%i:DATA:SDAT?' %( self._ci, trace))
            return np.array(data)[::2]
        else:
            raise ValueError('Keysight_VNA_E5080B_spectrum: Set the measurement class to the spectrum analyzer mode.')



# LMX2595
# Grigorev Sasha 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# Refer to the 'MS2690A/MS2691A/MS2692A or MS2830A Signal
# Analyzer Operation Manual (Mainframe Remote Control)' for detailed
# specifications on IEEE488.2 common device messages and application
# common device messages.


from qkit.core.instrument_base import Instrument
import zerorpc
import logging
import types


def set_bit(value, index, bit):
    mask = 1 << index
    value &= ~mask
    if bit:
        value |= mask
    return value


def get_bit(value, bit):
    return (value & (1 << bit)) != 0


def to_bytes(number, leng):
    return [(number >> (8 * i)) & 0xff for i in range(leng - 1, -1, -1)]


def from_bytes(byte):
    a = []
    for j in range(len(byte)):
        for i in range(8):
            a.append((byte[len(byte) - j - 1] & 1 << i) * 2 ** (8 * j))
    return sum(a)


class Lmx2595(Instrument):
    """
    This is the driver for the TI Lmx2595 Signal Genarator

    Usage:
    Initialize with
    <name> = instruments.create('<name>', 'Lmx2595', address='<Ethernet address>)
    """

    def __init__(self, name, address, reset=True):
        """
        Initializes the Lmx2595, and communicates with the wrapper.

        Input:
            name (string)    : name of the instrument
            address (string) : Ethernet address

        Output:
            True if initialization was successful
        """
        logging.info(__name__ + ' : Initializing instrument Lmx2595')
        Instrument.__init__(self, name, tags=['physical'])
        self.c = zerorpc.Client()
        self.c.connect(address)
        self.r = 2
        self.params = {
            'StA': [44, 6, 1],
            'StB': [44, 7, 1],
            'Vco_calib_st': [0, 3, 1],
            'Reset_St': [0, 1, 1],
            'PowerA': [44, 8, 6],
            'PowerB': [45, 0, 6],
            'Num_b1': [43, 0, 16],
            'Num_b0': [42, 0, 16],
            'N_dev': [36, 0, 16],
            'ChDevA_st': [45, 11, 2],
            'ChDevB_st': [46, 0, 2]
        } # Dict contains address of register, start bit and length
        self.vco_max = 15e9  # default values according to manual
        self.vco_min = 7.5e9

        self.add_parameter('frequency', type=float, flags=Instrument.FLAG_GETSET, minval=200e6, maxval=14.9e9)

        if reset:
            self.reset()

    def _set_parameter(self, name, value, mode='full'):
        """
        Function to set internal Lmx2595 parameters

        Input:
            name: the name of parameter according to dictionary
            value: the value of the parameter
            mode: 'full' or 'part' depending on opportunity to write whole register or only part of it

        Output:
            None

        """
        if mode == 'full':  # So we do not care what is now written in the register and can overwrite it
            self.c.write_data([self.params[name][0], value[0], value[1]]) #Write the bytes
        if mode == 'part':  # Here we need to overwrite only sertain bits of the register using the function, written at script on rpi
            reg = self.c.read_data(self.params[name][0]) #Read the register value
            if self.params[name][1] < 8:  # That means that the value should be changed is the part of first byte
                wr_reg = reg[2]
                index = self.params[name][1]
                leng = self.params[name][2]
                for i in range(index, index + leng):
                    wr_reg = set_bit(wr_reg, i, get_bit(value, i - index))
                self.c.write_data([self.params[name][0], reg[1], wr_reg])  # Write the changed data
            else:
                wr_reg = reg[1]
                index = self.params[name][1] - 8
                leng = self.params[name][2]
                for i in range(index, index + leng):
                    wr_reg = set_bit(wr_reg, i, get_bit(value, i - index))
                self.c.write_data([self.params[name][0], wr_reg, reg[2]])  # Write the changed data

    def get_value(self, name, mode='full'):
        """
        Function for obtaining values of parameters

        Input:
            name: name of parameter
            mode: 'full' or 'part' depending on opportunity to read whole register or only part of it

        Output:
            Value of the parameter
        """
        if mode == 'full':
            return self.c.read_data(self.params[name][0])  # Just read the whole register and convert data into numbers
        if mode == 'part':
            reg = self.c.read_data(self.params[name][0])
            if self.params[name][1] < 8:  # The value we are interested in is in the first byte
                rd_reg = reg[2]
                index = self.params[name][1]
                leng = self.params[name][2]
                res = []
                for i in range(index, index + leng):
                    res.append(get_bit(rd_reg, i) * 2 ** (i - index))  # Read only part of a certain byte
                return sum(res)
            else:
                rd_reg = reg[1]
                index = self.params[name][1] - 8
                leng = self.params[name][2]
                res = []
                for i in range(index, index + leng):
                    res.append(get_bit(rd_reg, i) * 2 ** (i - index))  # Read only part of a certain byte
                return sum(res)

    def get_all(self):
        """
        This is the function to read all the parameters of the device

        Input:
            None

        Output:
            None
        """
        logging.info(__name__ + ' : get all')
        self.get_frequency()
        self.get_power_chA()
        self.get_power_chB()
        self.get_status_chA()
        self.get_status_chB()

    def set_status_chA(self, status):
        """
        Function to set output A status

        Input:
            Status: True (1) - turn on, False(0) - turn off (inverted in function due to device issues).

        Output:
            None
        """
        self._set_parameter('StA', ~status, mode='part')

    def set_status_chB(self, status):
        """
        Function to set output B status

        Input:
            Status: True (1) - turn on, False(0) - turn off (inverted in function due to device issues).

        Output:
            None
        """
        self._set_parameter('StB', ~status, mode='part')

    def get_status_chA(self):
        """
        Function to get output A status

        Input:
            None

        Output:
            Status: True -turned on, False - turned off
        """
        return self.get_value('StA', mode='part') != 1

    def get_status_chB(self):
        """
        Function to get output B status

        Input:
            None

        Output:
            Status: True -turned on, False - turned off
        """
        return self.get_value('StB', mode='part') != 1

    def set_power_chB(self, power):
        """
        Function to set the power value on B output of the device

        Input:
            power: value of register from 0 to 31. (Bigger value provides more power)

        Output:
            True if setting was successful
        """
        self._set_parameter('PowerB', power, mode='part')
        return True

    def set_power_chA(self, power):
        """
        Function to set the power value on A output of the device

        Input:
            power: value of register from 0 to 31. (Bigger value provides more power)

        Output:
            True if setting was successful
        """
        self._set_parameter('PowerA', power, mode='part')
        return True

    def get_power_chB(self):
        """
        Function to get the power value on B output of the device

        Input:
            None

        Output:
            Power: Value of register from 0 to 31. (Bigger value provides more power).
        """
        return self.get_value('PowerB', mode='part')

    def get_power_chA(self):
        """
        Function to get the power value on A output of the device

        Input:
            None

        Output:
            Power: Value of register from 0 to 31. (Bigger value provides more power).
        """
        return self.get_value('PowerA', mode='part')

    def do_set_frequency(self, freq):
        """
        This is the function to set the frequency in Hz of both outputs of the device

        Input:
            freq - frequency to be set

        Output:
            True if setting was successful
        """
        self.r = self._set_mode(freq)
        q = freq % 1e8 / 1e5 * self.r * 10 // 10000
        z = freq % 1e8 / 1e5 * self.r % 1000
        x = int(freq / 1e8) * self.r + int(q)
        n_dev = to_bytes(x, 2)
        num = to_bytes(int(z), 4)
        self._set_parameter('N_dev', n_dev, mode='full')
        self._set_parameter('Num_b1', [num[2], num[3]])
        self._set_parameter('Num_b0', [num[0], num[1]])
        self._vco_calib_reset()
        return True

    def do_get_frequency(self):
        """
        This is the function to get frequency in Hz of both output of the device

        Input:
            None

        Output:
            Frequency value
        """
        n_dev = from_bytes(self.get_value('N_dev', mode='full'))
        num = from_bytes(self.get_value('Num_b0', mode='full')+self.get_value('Num_b1', mode='full'))
        return (n_dev*1e8+num*1e5)/self.r

    def _set_mode(self, freq):
        """
        Function to set status and value of frequency dividers depending on frequency value. Used only inside 'set_frequency'
        function

        Input:
            freq - Frequency to be set

        Output:
            Channel divider value
        """
        cond = []
        a = [1, 2, 4, 6, 8, 12, 16, 24, 32, 48]  # possible devider values
        for i in a:
            cond.append(self.vco_min / i <= freq< self.vco_max / i)
        for i in range(len(cond)):
            if cond[i] and a[i] == 1:
                self._set_parameter('ChDevA_st', 1, mode='part') #Turn off ChDev chA
                self._set_parameter('ChDevB_st', 1, mode='part') #Turn_off ChDev chB
                return a[i]
            elif cond[i]:
                self._set_chdev_val(a[i])					 #Set ChDev Value
                self._set_parameter('ChDevA_st', 0, mode='part') #Turn on ChDev chA
                self._set_parameter('ChDevB_st', 0, mode='part') #Turn on ChDev chB
                return a[i]

    def _set_chdev_val(self, a):
        """
        Sets certain byte sequence for ChDev value is needed only use in setting frequency

        Input:
            a - ch_dev value

        Output:
            None
        """
        chd_dict_mask = {
            2: [0x4b, 0x08, 0x00],
            4: [0x4b, 0x08, 0x40],
            6: [0x4b, 0x08, 0x80],
            8: [0x4b, 0x08, 0xc0],
            12: [0x4b, 0x09, 0x00],
            16: [0x4b, 0x09, 0x40],
            24: [0x4b, 0x09, 0x80],
            32: [0x4b, 0x09, 0xC0],
            48: [0x4b, 0x0A, 0x00]
        }
        self.c.write_data(chd_dict_mask[a])

    def _vco_calib_reset(self):
        """
        This is function to reset Voltage controlled oscillator's calibration (used only in setting frequency)

        Input:
            None

        Output:
            None
        """
        self._set_parameter('Vco_calib_st', 1, mode='part')
        self._set_parameter('Vco_calib_st', 0, mode='part')

    def reset(self):
        """
        Function to load default value for all registers

        Input:
            None

        Output:
            None
        """

        self._set_parameter('Reset_St', 1, mode='part')
        self._set_parameter('Reset_St', 0, mode='part')
        default = {
            'R112':	0x700000,
            'R111':	0x6F0000,
            'R110':	0x6E0000,
            'R109':	0x6D0000,
            'R108':	0x6C0000,
            'R107':	0x6B0000,
            'R106':	0x6A0000,
            'R105':	0x690021,
            'R104':	0x680000,
            'R103':	0x670000,
            'R102':	0x660000,
            'R101':	0x650011,
            'R100':	0x640000,
            'R99':	0x630000,
            'R98':	0x620000,
            'R97':	0x610888,
            'R96':	0x600000,
            'R95':	0x5F0000,
            'R94':	0x5E0000,
            'R93':	0x5D0000,
            'R92':	0x5C0000,
            'R91':	0x5B0000,
            'R90':	0x5A0000,
            'R89':	0x590000,
            'R88':	0x580000,
            'R87':	0x570000,
            'R86':	0x560000,
            'R85':	0x550000,
            'R84':	0x540000,
            'R83':	0x530000,
            'R82':	0x520000,
            'R81':	0x510000,
            'R80':	0x500000,
            'R79':	0x4F0000,
            'R78':	0x4E0003,
            'R77':	0x4D0000,
            'R76':	0x4C000C,
            'R75':	0x4B0800,
            'R74':	0x4A0000,
            'R73':	0x49003F,
            'R72':	0x480001,
            'R71':	0x470081,
            'R70':	0x46C350,
            'R69':	0x450000,
            'R68':	0x4403E8,
            'R67':	0x430000,
            'R66':	0x4201F4,
            'R65':	0x410000,
            'R64':	0x401388,
            'R63':	0x3F0000,
            'R62':	0x3E0322,
            'R61':	0x3D00A8,
            'R60':	0x3C0000,
            'R59':	0x3B0001,
            'R58':	0x3A8001,
            'R57':	0x390020,
            'R56':	0x380000,
            'R55':	0x370000,
            'R54':	0x360000,
            'R53':	0x350000,
            'R52':	0x340820,
            'R51':	0x330080,
            'R50':	0x320000,
            'R49':	0x314180,
            'R48':	0x300300,
            'R47':	0x2F0300,
            'R46':	0x2E07FC,
            'R45':	0x2DC0DF,
            'R44':	0x2C1FE3,
            'R43':	0x2B0000,
            'R42':	0x2A0000,
            'R41':	0x290000,
            'R40':	0x280000,
            'R39':	0x2703E8,
            'R38':	0x260000,
            'R37':	0x250404,
            'R36':	0x24008C,
            'R35':	0x230004,
            'R34':	0x220000,
            'R33':	0x211E21,
            'R32':	0x200393,
            'R31':	0x1F43EC,
            'R30':	0x1E318C,
            'R29':	0x1D318C,
            'R28':	0x1C0488,
            'R27':	0x1B0002,
            'R26':	0x1A0DB0,
            'R25':	0x190624,
            'R24':	0x18071A,
            'R23':	0x17007C,
            'R22':	0x160001,
            'R21':	0x150401,
            'R20':	0x14E048,
            'R19':	0x1327B7,
            'R18':	0x120064,
            'R17':	0x11012C,
            'R16':	0x100080,
            'R15':	0x0F064F,
            'R14':	0x0E1E70,
            'R13':	0x0D4000,
            'R12':	0x0C5001,
            'R11':	0x0B0018,
            'R10':	0x0A10D8,
            'R9':	0x090604,
            'R8':	0x082000,
            'R7':	0x0740B2,
            'R6':	0x06C802,
            'R5':	0x0500C8,
            'R4':	0x040A43,
            'R3':	0x030642,
            'R2':	0x020500,
            'R1':	0x010808,
            'R0':	0x002518
            }
        num = []
        for i in range(112, -1, -1):
            num.append(default['R'+str(i)])
        bw = [to_bytes(i, 3) for i in num]
        for i in range(len(bw)):
            self.c.write_data(bw[i])
